"Namespace definitions"


# XML namespaces
XML_SCHEMA = "http://www.w3.org/2001/XMLSchema"
XML_SCHEMA_INSTANCE = "http://www.w3.org/2001/XMLSchema-instance"

# SOAP namespaces
SOAP_ENV = "http://www.w3.org/2003/05/soap-envelope"

# Web services security namespaces
WSS_PWDIGEST = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordDigest"
WSS_BASE64BIN = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
WSS_SECUTIL = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"
WSS_SECEXT = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"

# Onvif base namespaces for schema, device, media & ptz
OVF_SCHEMA = "http://www.onvif.org/ver10/schema"
OVF_DEVICE = "http://www.onvif.org/ver10/device/wsdl"
OVF_MEDIA = "http://www.onvif.org/ver10/media/wsdl"
OVF_PTZ = "http://www.onvif.org/ver20/ptz/wsdl"

# Onvif zoom spaces
OVF_ZS_VGS = "http://www.onvif.org/ver10/tptz/ZoomSpaces/VelocityGenericSpace"
OVF_ZS_TGS = "http://www.onvif.org/ver10/tptz/ZoomSpaces/TranslationGenericSpace"
OVF_ZS_PGS = "http://www.onvif.org/ver10/tptz/ZoomSpaces/PositionGenericSpace"
OVF_ZS_ZGSS = "http://www.onvif.org/ver10/tptz/ZoomSpaces/ZoomGenericSpeedSpace"

# Onvif pan & tilt spaces
OVF_PTS_VGS = "http://www.onvif.org/ver10/tptz/PanTiltSpaces/VelocityGenericSpace"
OVF_PTS_TGS = "http://www.onvif.org/ver10/tptz/PanTiltSpaces/TranslationGenericSpace"
OVF_PTS_PGS = "http://www.onvif.org/ver10/tptz/PanTiltSpaces/PositionGenericSpace"
OVF_PTS_GSS = "http://www.onvif.org/ver10/tptz/PanTiltSpaces/GenericSpeedSpace"



import re, ast, logging, argparse, os
from xml.dom import minidom
from . import messages
from .pyonvif import NoCameraFoundException, OnvifCam


log = logging.getLogger(__name__)


def get_commands():
    "return list of available commands & their params"

    xpr = re.compile("{[a-z_]+}")
    cmds = {}
    with open(messages.__file__) as msgsf:
       msgs = ast.parse(msgsf.read())
       for i in msgs.body:
          if isinstance(i, ast.Assign):
             cmd = i.targets[0].id
             if cmd.startswith('_'):
                 continue
             parms = [p.strip('{}') for p in re.findall(xpr, i.value.s)] or []
             cmds[cmd] = parms
    return cmds


def command():
   "the command-line client"

   parser = argparse.ArgumentParser()
   parser.add_argument("-a", "--address", help="camera address (and optionally :port)")
   parser.add_argument("-s", "--servicepath", help="service path", default="/onvif/device_service")
   parser.add_argument("-p", "--profile", help="profile")
   parser.add_argument("-u", "--user", help="username", default=None)
   parser.add_argument("-v", "--verbose", action='store_true', help="Log level")

   subparsers = parser.add_subparsers(title='onvif commands',
                                      description='valid commands are listed below',
                                      help='enter command name and parameters',
                                      dest="_cmd")

   cmds = get_commands()
   for cmd, parms in cmds.items():
      hlp = cmd.lower().replace('_', ' ').capitalize()
      cmdparser = subparsers.add_parser(cmd, help=hlp)
      for parm in parms:
         cmdparser.add_argument(parm, type=str, help=parm + " argument")

   args = parser.parse_args()

   if not args._cmd:
      parser.print_help()
      parser.exit()

   if args.verbose:
      logging.basicConfig(level=logging.DEBUG)

   pwd = None
   if args.user:
      pwd = os.environ.get("CAMERA_PASSWORD")
      if not pwd:
         parser.exit(message="no password; please add it as CAMERA_PASSWORD to shell environment\n")

   try:
      addr, port = args.address.split(':')
   except:
      addr = args.address
      port = 80

   try:
      c = OnvifCam(addr=addr, port=port, pth=args.servicepath, prf=args.profile, usr=args.user, pwd=pwd,\
                   verbose=args.verbose)
   except NoCameraFoundException:
      parser.exit(message="no camera found\n")

   result = c.execute(args._cmd, **dict(args._get_kwargs()))
   if result:
      tree = minidom.parseString(result)
      body = tree.getElementsByTagName("SOAP-ENV:Body")[0]
      for n in body.childNodes:
         print(n.toprettyxml())
      return




"""Onvif SOAP message definitions"""


# System & service requests


_SOAP_ENV = """
<s:Envelope xmlns:s="{SOAP_ENV}">{content}</s:Envelope>
"""

_SOAP_BODY = """
<s:Body xmlns:xsi="{XML_SCHEMA_INSTANCE}" xmlns:xsd="{XML_SCHEMA}">{content}</s:Body>
"""

_AUTH_HEADER = """
<s:Header>
    <Security s:mustUnderstand="1" xmlns="{WSS_SECEXT}">
        <UsernameToken>
            <Username>{username}</Username>
            <Password Type="{WSS_PWDIGEST}">{pdigest}</Password>
            <Nonce EncodingType="{WSS_BASE64BIN}">{nonce}</Nonce>
            <Created xmlns="{WSS_SECUTIL}">{created}</Created>
        </UsernameToken>
    </Security>
</s:Header>
"""

GET_SYSTEM_DATETIME = """
<GetSystemDateAndTime xmlns="{OVF_DEVICE}"/>
"""

GET_CAPABILITIES = """
<GetCapabilities xmlns="{OVF_DEVICE}"><Category>All</Category></GetCapabilities>
"""

GET_SERVICE_CAPABILITIES = """
<GetServiceCapabilities xmlns="{OVF_DEVICE}"></GetServiceCapabilities>
"""

GET_SERVICES = """
<GetServices xmlns="{OVF_DEVICE}"><IncludeCapability>false</IncludeCapability></GetServices>
"""

GET_PROFILES = """
<GetProfiles xmlns="{OVF_MEDIA}"></GetProfiles>
"""

GET_DEVICE_INFO = """
<GetDeviceInformation xmlns="{OVF_DEVICE}"></GetDeviceInformation>
"""

GET_NODE = """
<GetNode xmlns="{OVF_PTZ}">
    <NodeToken>{node}</NodeToken>
</GetNode>
"""


# Movements

RELATIVE_MOVE = """
<RelativeMove xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
    <Translation>
        <PanTilt x="{x}" y="{y}" space="{OVF_PTS_TGS}" xmlns="{OVF_SCHEMA}"/>
    </Translation>
    <Speed>
        <PanTilt x="{xspeed}" y="{yspeed}" space="{OVF_PTS_GSS}" xmlns="{OVF_SCHEMA}"/>
    </Speed>
</RelativeMove>
"""

RELATIVE_MOVE_ZOOM = """
<RelativeMove xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
    <Translation>
        <Zoom x="{z}" space="{OVF_ZS_TGS}" xmlns="{OVF_SCHEMA}"/>
    </Translation>
    <Speed>
        <Zoom x="{zspeed}" space="{OVF_ZS_ZGSS}" xmlns="{OVF_SCHEMA}"/>
    </Speed>
</RelativeMove>
"""

ABSOLUTE_MOVE = """
<AbsoluteMove xmlns="{OVF_PTZ}">
    <Position>
        <ProfileToken>{profile}</ProfileToken>
        <PanTilt x="{x}" y="{y}" space="{OVF_PTS_PGS}" xmlns="{OVF_SCHEMA}"/>
        <Zoom x="{z}" space="{OVF_ZS_PGS}" xmlns="{OVF_SCHEMA}"/>
    </Position>
</AbsoluteMove>
"""

CONTINUOUS_MOVE = """
<ContinuousMove xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
    <Velocity>
        <PanTilt x="{x}" y="{y}" space="{OVF_PTS_VGS}" xmlns="{OVF_SCHEMA}"/>
    </Velocity>
</ContinuousMove>
"""

STOP = """
<Stop xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
    <PanTilt>{ptstop}</PanTilt>
    <Zoom>{zstop}</Zoom>
</Stop>
"""

CONTINUOUS_MOVE_ZOOM = """
<ContinuousMove xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
    <Velocity>
        <Zoom x="{z}" space="{OVF_ZS_VGS}" xmlns="{OVF_SCHEMA}"/>
    </Velocity>
</ContinuousMove>
"""


# Preset use and manipulation

SET_PRESET = """
    <SetPreset xmlns="{OVF_PTZ}">
        <ProfileToken>{profile}</ProfileToken>
        <PresetName>{preset}</PresetName>
    </SetPreset>
"""

GET_PRESETS = """
<GetPresets xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
</GetPresets>
"""

GOTO_PRESET = """
<GotoPreset xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
    <PresetToken>{preset}</PresetToken>
    <Speed>
        <PanTilt x="{xspeed}" y="{yspeed}" xmlns="{OVF_SCHEMA}"/>
        <Zoom x="{zspeed}" xmlns="{OVF_SCHEMA}"/>
    </Speed>
</GotoPreset>
"""

REMOVE_PRESET = """
<RemovePreset xmlns="{OVF_PTZ}">
    <ProfileToken>{profile}</ProfileToken>
    <PresetToken>{preset}</PresetToken>
</RemovePreset>
"""


# Video sources

GET_VIDEO_SOURCES = """
<GetVideoSources xmlns="{OVF_MEDIA}"/>
"""


# Video stream URI

GET_STREAM_URI = """
<GetStreamUri xmlns="{OVF_MEDIA}">
    <StreamSetup>
        <Stream xmlns="{OVF_SCHEMA}">RTP-Unicast</Stream>
        <Transport xmlns="{OVF_SCHEMA}">
            <Protocol>UDP</Protocol>
        </Transport>
    </StreamSetup>
    <ProfileToken>{profile}</ProfileToken>
</GetStreamUri>"""



from setuptools import setup


def get_file(fn):
    with open(fn) as fp:
        return fp.read()

setup(
    name='pyOnvif',
    version="1.1",
    description='Simple Onvif camera client',
    long_description=(
        get_file('README.rst') + '\n\n' + get_file('HISTORY.rst')
    ),
    author="Pekka Jäppinen",
    maintainer="Petri Savolainen",
    url='https://github.com/koodaamo/pyOnvif',
    install_requires=[],
    extras_require={
        'discovery':  ["WSDiscovery"]
    },
    packages=[
        'pyonvif',
    ],
    include_package_data=True,
    license="GPLv3",
    zip_safe=False,
    keywords='onvif',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points = {'console_scripts': ['pyonvif=pyonvif.cmdline:command']}
)



#!/usr/bin/env python3
## -*- coding: utf-8 -*-
## Onvif camera controlling class that uses soap and wsdiscovery
## Pekka Jäppinen 2014
## Conversion to Python3 Petri Savolainen 2017

import datetime
import base64
from hashlib import sha1
import argparse
import logging
from random import SystemRandom
import string
import http.client
from urllib.parse import urlparse
from . import messages, namespaces
from wsdiscovery import WSDiscovery


logging.basicConfig()
logger = logging.getLogger("onvifcam")
logger.setLevel(logging.INFO)

SPATH = "/onvif/device_service"
PROF = "prof0"
pool = string.ascii_letters + string.digits + string.punctuation


nsmap = {k:v for k, v in namespaces.__dict__.items() if not k.startswith('__')}


class NoCameraFoundException(Exception):
   ""

class OnvifCam():

   def __init__(self, addr=None, port=80, pth=SPATH, prf=PROF, usr=None, pwd=None, basicauth=False, verbose=None):

      if verbose:
         logger.setLevel(logging.DEBUG)

      self.x = 0
      self.movespeed = "0.5"
      self.zoomspeed = "0.5"
      self.profiles = []
      self.profile = prf
      self.username, self.password = usr, pwd
      if usr and pwd:
         masked = len(pwd) * '*'
         logger.info(f"authorization credentials: {usr}:{masked}")
      else:
         logger.info("no authorization credentials given")
      self.address = addr
      self.port = port
      self.path = pth
      self.profile = prf
      self.basicauth = basicauth

      if not addr:
         logger.info("attempting to discover camera")
         found = self.discover()
         if not found:
            raise NoCameraFoundException("no services discovered")
         else:
            self.address, self.port, self.path = found

      logger.info(f"attempting to connect to camera at {self.address}:{self.port}")
      self.connection = http.client.HTTPConnection(self.address, port=self.port)


   def execute(self, command, **parms):
      tmpl = getattr(messages, command)
      parms.update(namespaces.__dict__)
      try:
         result = tmpl.format(**parms)
      except KeyError as exc:
         logger.error("not making Onvif {command}: missing parameter: {exc.args[0]}")
      else:
         response = self.sendSoapMsg(result)
         return response


   def discover(self, attempts=3):
      "Use WS DIscovery to try & find available camera(s) a number of times"

      wsd = WSDiscovery()
      wsd.start()

      attempts_left = attempts
      found = None

      while not found and attempts_left:

         found = wsd.searchServices()
         for service in found:
            url = urlparse(service.getXAddrs()[0])
            if "onvif" not in url.path:
               logger.warning("not an Onvif service path: %s", url.path)
            else:
               found = url
               logger.info(f"discovered onvif service at {found.netloc}{found.path}")
               break

         attempts_left -= 1

      return (found.hostname, found.port or 80, found.path) if found else None

   def sendSoapMsg(self, bmsg):
      body = messages._SOAP_BODY.format(content=bmsg, **nsmap)
      hdrs = {}
      if self.username and self.password:
         if self.basicauth:
            auth_plaintext = f"{self.username}:{self.password}".encode("ascii")
            auth_encoded = base64.b64encode(auth_plaintext).decode("ascii")
            hdrs = {"Authorization": "Basic " + auth_encoded}
         else:
            body = self.getAuthHeader() + body
      soapmsg = messages._SOAP_ENV.format(content=body, **nsmap)
      if not self.connection:
         self.connection = http.client.HTTPConnection(self.address)

      try:
         self.connection.request("POST", self.path, soapmsg, headers=hdrs)
      except ConnectionRefusedError:
         logger.error("cannot connect")
         return None

      logger.debug(soapmsg)

      resp = self.connection.getresponse()
      if resp.status != 200:
         logger.error("failure: %s %s", resp.status, resp.reason)
      else:
         return resp.read()

   def getAuthHeader(self):
      created = datetime.datetime.now().isoformat().split(".")[0]
      n64 = ''.join(SystemRandom().choice(pool) for _ in range(22))
      nonce = base64.b64encode(n64.encode('ascii')).decode("ascii")
      base = (n64 + created + self.password).encode("ascii")
      pdigest= base64.b64encode(sha1(base).digest()).decode("ascii")
      parms = {}
      parms.update(**nsmap)
      username = self.username
      parms.update(**locals())
      return messages._AUTH_HEADER.format(**parms)



